/* src/ihw.c - low level control functions
 *
 * --------------------------------------------------------------------
 *
 * Copyright (C) 2003  ACX100 Open Source Project
 *
 *   The contents of this file are subject to the Mozilla Public
 *   License Version 1.1 (the "License"); you may not use this file
 *   except in compliance with the License. You may obtain a copy of
 *   the License at http://www.mozilla.org/MPL/
 *
 *   Software distributed under the License is distributed on an "AS
 *   IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 *   implied. See the License for the specific language governing
 *   rights and limitations under the License.
 *
 *   Alternatively, the contents of this file may be used under the
 *   terms of the GNU Public License version 2 (the "GPL"), in which
 *   case the provisions of the GPL are applicable instead of the
 *   above.  If you wish to allow the use of your version of this file
 *   only under the terms of the GPL and not to allow others to use
 *   your version of this file under the MPL, indicate your decision
 *   by deleting the provisions above and replace them with the notice
 *   and other provisions required by the GPL.  If you do not delete
 *   the provisions above, a recipient may use your version of this
 *   file under either the MPL or the GPL.
 *
 * --------------------------------------------------------------------
 *
 * Inquiries regarding the ACX100 Open Source Project can be
 * made directly to:
 *
 * acx100-users@lists.sf.net
 * http://acx100.sf.net
 *
 * --------------------------------------------------------------------
 */

#include <linux/config.h>
#include <linux/version.h>

#include <linux/module.h>
#include <linux/kernel.h>

#if (WLAN_HOSTIF==WLAN_USB)
#include <linux/usb.h>
#endif

#include <linux/if_arp.h>
#include <linux/wireless.h>
#if WIRELESS_EXT > 12
#include <net/iw_handler.h>
#endif

#include <wlan_compat.h>

#include <linux/pci.h>

#include <linux/etherdevice.h>


/*================================================================*/
/* Project Includes */

#include <p80211hdr.h>
#include <acx100.h>
#include <acx100_helper.h>
#include <ihw.h>

extern void acx100_dump_bytes(void *,int);

#if (WLAN_HOSTIF==WLAN_USB)
/* try to make it compile for both 2.4.x and 2.6.x kernels */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 0)

static inline int submit_urb(struct urb *urb, int mem_flags) {
	        return usb_submit_urb(urb, mem_flags);
}

static void acx100usb_control_complete(struct urb *urb)
{
	FN_ENTER;
	FN_EXIT(0,OK);
}

#else

/* 2.4.x kernels */
#define USB_24	1

static inline int submit_urb(struct urb *urb, int mem_flags) {
	        return usb_submit_urb(urb);
}

static void acx100usb_control_complete(struct urb *urb, struct pt_regs *regs)
{
	FN_ENTER;
	FN_EXIT(0,OK);
}

#endif

#define FILL_SETUP_PACKET(_pack,_rtype,_req,_val,_ind,_len) \
								(_pack)[0]=_rtype; \
								(_pack)[1]=_req; \
								((unsigned short *)(_pack))[1]=_val; \
								((unsigned short *)(_pack))[2]=_ind; \
								((unsigned short *)(_pack))[3]=_len;

#endif



/*****************************************************************************
 *
 * The Really Low Level Stuff (TM)
 *
 ****************************************************************************/

#ifndef IO_AS_MACROS
/*----------------------------------------------------------------
* acx100_read_reg32
*
*
* Arguments:
*
* Returns:
*
* Side effects:
*
* Call context:
*
* STATUS: FINISHED
*
* Comment:
*
*----------------------------------------------------------------*/
inline UINT32 acx100_read_reg32(wlandevice_t *priv, UINT offset)
{
#if ACX_IO_WIDTH == 32
	return readl(priv->iobase + offset);
#else 
	return readw(priv->iobase + offset)
	    + (readw(priv->iobase + offset + 2) << 16);
#endif
}

/*----------------------------------------------------------------
* acx100_read_reg16
*
*
* Arguments:
*
* Returns:
*
* Side effects:
*
* Call context:
*
* STATUS: FINISHED
*
* Comment:
*
*----------------------------------------------------------------*/
inline UINT16 acx100_read_reg16(wlandevice_t *priv, UINT offset)
{
	return readw(priv->iobase + offset);
}

/*----------------------------------------------------------------
* acx100_read_reg8
*
*
* Arguments:
*
* Returns:
*
* Side effects:
*
* Call context:
*
* STATUS: FINISHED
*
* Comment:
*
*----------------------------------------------------------------*/
inline UINT8 acx100_read_reg8(wlandevice_t *priv, UINT offset)
{
	return readb(priv->iobase + offset);
}

/*----------------------------------------------------------------
* acx100_write_reg32
*
*
* Arguments:
*
* Returns:
*
* Side effects:
*
* Call context:
*
* STATUS: FINISHED
*
* Comment:
*
*----------------------------------------------------------------*/
inline void acx100_write_reg32(wlandevice_t *priv, UINT offset, UINT val)
{
#if ACX_IO_WIDTH == 32
	writel(val, priv->iobase + offset);
#else 
	writew(val & 0xffff, priv->iobase + offset);
	writew(val >> 16, priv->iobase + offset + 2);
#endif
}

/*----------------------------------------------------------------
* acx100_write_reg16
*
*
* Arguments:
*
* Returns:
*
* Side effects:
*
* Call context:
*
* STATUS: FINISHED
*
* Comment:
*
*----------------------------------------------------------------*/
inline void acx100_write_reg16(wlandevice_t *priv, UINT offset, UINT16 val)
{
	writew(val, priv->iobase + offset);
}

/*----------------------------------------------------------------
* acx100_write_reg8
*
*
* Arguments:
*
* Returns:
*
* Side effects:
*
* Call context:
*
* STATUS: FINISHED
*
* Comment:
*
*----------------------------------------------------------------*/
inline void acx100_write_reg8(wlandevice_t *priv, UINT offset, UINT val)
{
	writeb(val, priv->iobase + offset);
}
#endif

/*****************************************************************************
 * 
 * Intermediate Level
 *
 ****************************************************************************/

/* Info mailbox format:
2 bytes: type
2 bytes: status
more bytes may follow
    docs say about status:
	0x0000 info available (set by hw)
	0x0001 information received (must be set by host)
	0x1000 info available, mailbox overflowed (messages lost) (set by hw)
    but in practice we've seen:
	0x9000 when we did not set status to 0x0001 on prev message
	0x1001 when we did set it
	0x0000 was never seen
    conclusion: this is really a bitfield:
    0x1000 is 'info available' bit
    'mailbox overflowed' bit is 0x8000, not 0x1000
    value of 0x0000 probably means that there is no message at all
    P.S. I dunno how in hell hw is supposed to notice that messages are lost -
    it does NOT clear bit 0x0001, and this bit will probably stay forever set
    after we set it once. Let's hope this will be fixed in firmware someday
*/

void acx100_get_info_state(wlandevice_t *priv)
{
	UINT32 value;

	acx100_write_reg32(priv, priv->io[IO_ACX_SLV_END_CTL], 0x0);
	acx100_write_reg32(priv, priv->io[IO_ACX_SLV_MEM_CTL], 0x1);

	acx100_write_reg32(priv, priv->io[IO_ACX_SLV_MEM_ADDR],
		acx100_read_reg32(priv, priv->io[IO_ACX_INFO_MAILBOX_OFFS]));

	value = acx100_read_reg32(priv, priv->io[IO_ACX_SLV_MEM_DATA]);

	priv->info_type = value & 0xffff;
	priv->info_status = value >> 16;

	/* inform hw that we have read this info message */
	acx100_write_reg32(priv, priv->io[IO_ACX_SLV_MEM_DATA], priv->info_type | 0x00010000);
	acx100_write_reg16(priv, priv->io[IO_ACX_INT_TRIG], INT_TRIG_ACK); /* now bother hw to notice it */

	acxlog(L_CTL, "info_type 0x%04x, info_status 0x%04x\n", priv->info_type, priv->info_status);
}

/*----------------------------------------------------------------
* acx100_get_cmd_state
*
*
* Arguments:
*
* Returns:
*
* Side effects:
*
* Call context:
*
* STATUS: FINISHED
*
* Comment:
*
*----------------------------------------------------------------*/
void acx100_get_cmd_state(wlandevice_t *priv)
{
	UINT32 value;

	acx100_write_reg32(priv, priv->io[IO_ACX_SLV_END_CTL], 0x0);
	acx100_write_reg32(priv, priv->io[IO_ACX_SLV_MEM_CTL], 0x1); /* why auto increment?? */

	acx100_write_reg32(priv, priv->io[IO_ACX_SLV_MEM_ADDR], 
		acx100_read_reg32(priv, priv->io[IO_ACX_CMD_MAILBOX_OFFS]));

	value = acx100_read_reg32(priv, priv->io[IO_ACX_SLV_MEM_DATA]);
	priv->cmd_type = (UINT16)value;
	priv->cmd_status = (UINT16)(value >> 16);

	acxlog(L_CTL, "cmd_type 0x%04x, cmd_status 0x%04x\n", priv->cmd_type, priv->cmd_status);
}

/*----------------------------------------------------------------
* acx100_write_cmd_type_or_status
*
*
* Arguments:
*
* Returns:
*
* Side effects:
*
* Call context:
*
* STATUS: FINISHED
*
* Comment:
*
*----------------------------------------------------------------*/
void acx100_write_cmd_type_or_status(wlandevice_t *priv, UINT val, INT is_status)
{
	acx100_write_reg32(priv, priv->io[IO_ACX_SLV_END_CTL], 0x0);
	acx100_write_reg32(priv, priv->io[IO_ACX_SLV_MEM_CTL], 0x1);

	acx100_write_reg32(priv, priv->io[IO_ACX_SLV_MEM_ADDR], 
		acx100_read_reg32(priv, priv->io[IO_ACX_CMD_MAILBOX_OFFS]));

	if (is_status)
		val <<= 16;
	acx100_write_reg32(priv, priv->io[IO_ACX_SLV_MEM_DATA], val);
}

/*----------------------------------------------------------------
* acx100_write_cmd_param
*
*
* Arguments:
*
* Returns:
*
* Side effects:
*
* Call context:
*
* STATUS: FINISHED
*
* Comment:
*
*----------------------------------------------------------------*/
static inline void acx100_write_cmd_param(wlandevice_t *priv, memmap_t *cmd, int len)
{
	memcpy(priv->CommandParameters, cmd, len);
}

/*----------------------------------------------------------------
* acx100_read_cmd_param
*
*
* Arguments:
*
* Returns:
*
* Side effects:
*
* Call context:
*
* STATUS: FINISHED
*
* Comment:
*
*----------------------------------------------------------------*/
static inline void acx100_read_cmd_param(wlandevice_t *priv, memmap_t *cmd, int len)
{
	memcpy(cmd, priv->CommandParameters, len);
}

static const char * const cmd_error_strings[] = {
	"Idle[TIMEOUT]",
	"[SUCCESS]",
	"Unknown Command",
	"Invalid Information Element",
	"unknown_FIXME",
	"channel invalid in current regulatory domain",
	"unknown_FIXME2",
	"Command rejected (read-only information element)",
	"Command rejected",
	"Already asleep",
	"Tx in progress",
	"Already awake",
	"Write only",
	"Rx in progress",
	"Invalid parameter",
	"Scan in progress"
};

/*----------------------------------------------------------------
* acx100_issue_cmd
* Excecutes a command in the command mailbox 
*
* Arguments:
*   *pcmdparam = an pointer to the data. The data mustn't include
*                the 4 byte command header!
*
* Returns:
*
* Side effects:
*
* Call context:
*
* STATUS: FINISHED
*
* Comment:
*
*----------------------------------------------------------------*/
#if (WLAN_HOSTIF!=WLAN_USB)
int acx100_issue_cmd(wlandevice_t *priv, UINT cmd,
			/*@null@*/ void *pcmdparam, int paramlen, UINT32 timeout)
{
	int counter;
	int result = NOT_OK;
	UINT16 irqtype = 0;
	UINT16 cmd_status;

	FN_ENTER;
	acxlog(L_CTL, "%s cmd 0x%X timeout %d.\n", __func__, cmd, timeout);

	if (!(priv->dev_state_mask & ACX_STATE_FW_LOADED))
	{
		acxlog(L_CTL, "firmware not loaded yet, cannot execute command!!\n");
		goto done;
	}
	
	if (cmd != ACX1xx_CMD_INTERROGATE) {
		acxlog(L_DEBUG,"input pdr (len=%d):\n",paramlen);
		acx100_dump_bytes(pcmdparam, paramlen);
	}

	/*** wait for ACX100 to become idle for our command submission ***/
	for (counter = 2000; counter > 0; counter--) {
		/* Auwtsh, busy-waiting!
		 * Hmm, yeah, but this function is probably supposed to be
		 * finished ASAP, so it's probably not such a terribly good
		 * idea to schedule away instead of busy-wait.
		 * Maybe let's just keep it as is. */
		acx100_get_cmd_state(priv);
		/* Test for IDLE state */
		if (!priv->cmd_status) 
			break;
	}

	/*** The card doesn't get IDLE, we're in trouble ***/
	if (counter == 0) {
		/* uh oh, cmd status still set, now we're in trouble */
		acxlog((L_BINDEBUG | L_CTL),
		       "Trying to issue a command to the ACX100 but the Command Register is not IDLE (%xh)\n", priv->cmd_status);
		goto done;
	}

	if(priv->irq_status & HOST_INT_CMD_COMPLETE) {
		acxlog((L_BINDEBUG | L_CTL), "irq status var is not empty: 0x%X\n", priv->irq_status);
		/* goto done; */
		priv->irq_status ^= HOST_INT_CMD_COMPLETE;
		acxlog((L_BINDEBUG | L_CTL), "irq status fixed to: 0x%X\n", priv->irq_status);
	}

	/*** now write the parameters of the command if needed ***/
	if (pcmdparam != NULL && paramlen != 0) {
		/* if it's an INTERROGATE command, just pass the length
		 * of parameters to read, as data */
		acx100_write_cmd_param(priv, pcmdparam,
			(cmd == ACX1xx_CMD_INTERROGATE) ? 0x4 : paramlen);
	}

	/*** now write the actual command type ***/
	priv->cmd_type = cmd;
	acx100_write_cmd_type_or_status(priv, cmd, 0);
	
	/*** execute command ***/
	acx100_write_reg16(priv, priv->io[IO_ACX_INT_TRIG],
			  acx100_read_reg16(priv, priv->io[IO_ACX_INT_TRIG]) | 0x01);

	/*** wait for IRQ to occur, then ACK it? ***/

	/*** make sure we have at least *some* timeout value ***/
	if (timeout == 0) {
		timeout = 1;
	}
	if (timeout > 120000) {
		timeout = 120000;
	}
	timeout *= 20;
	for (counter = 0; counter < timeout; counter++) {
		/* it's a busy wait loop, but we're supposed to be
		 * fast here, so better don't schedule away here?
		 * In theory, yes, but the timeout can be HUGE,
		 * so better schedule away sometimes */
		if (!priv->irqs_active && (irqtype =
		     acx100_read_reg16(priv, priv->io[IO_ACX_IRQ_STATUS_NON_DES])) & HOST_INT_CMD_COMPLETE) {

			acx100_write_reg16(priv, priv->io[IO_ACX_IRQ_ACK], HOST_INT_CMD_COMPLETE);
			break;
		} 
		if(priv->irqs_active && (irqtype = priv->irq_status & HOST_INT_CMD_COMPLETE)) {
			priv->irq_status ^= HOST_INT_CMD_COMPLETE;
			break;
		}

		if (counter % 30000 == 0)
		{
			acx100_schedule(HZ / 50);
		}
	}

	/*** Save state for debugging ***/
	acx100_get_cmd_state(priv);
	cmd_status = priv->cmd_status;
	
	/*** Put the card in IDLE state ***/
	priv->cmd_status = 0;
	acx100_write_cmd_type_or_status(priv, 0, 1);

	if (!(irqtype & HOST_INT_CMD_COMPLETE)) {
		acxlog(L_CTL,
			"Polling for an IRQ failed with %X, cmd_status %d, irqs_active %d, irq_status %X. Bailing.\n",
			irqtype, cmd_status, priv->irqs_active, priv->irq_status);
		if (debug & L_CTL)
			dump_stack();
		goto done;
	}

	if (1 != cmd_status) {
		acxlog(L_STD | L_CTL, "%s failed: %s [%d uSec] Cmd: %Xh, Result: %Xh\n",
				__func__,
				cmd_status <= 0x0f ?
				cmd_error_strings[cmd_status] : "UNKNOWN REASON",
				(timeout - counter) * 50,
				cmd,
				cmd_status);
		if (debug & L_CTL)
			dump_stack();

		/* zero out result buffer */
		if (pcmdparam != NULL && paramlen != 0) {
			memset(pcmdparam, 0, paramlen);
		}
	} else	{
		/*** read in result parameters if needed ***/
		if (pcmdparam != NULL && paramlen != 0) {
			if (cmd == ACX1xx_CMD_INTERROGATE) {
				acx100_read_cmd_param(priv, pcmdparam, paramlen);
        			acxlog(L_DEBUG,"output pdr (len=%d):\n",paramlen);
        			acx100_dump_bytes(pcmdparam, paramlen);
			}
		}
		result = OK;
	}

done:
	FN_EXIT(1, result);
	return result;
}
#else
int acx100_issue_cmd(wlandevice_t *priv,UINT cmd,void *pdr,int paramlen,UINT32 timeout) {
	int result,skipridheader,blocklen,inpipe,outpipe,acklen=sizeof(priv->usbin);
	int ucode;
	struct usb_device *usbdev;

	FN_ENTER;
	acxlog(L_CTL, "%s cmd 0x%X timeout %d.\n", __func__, cmd, timeout);

	skipridheader=0;
	/* ----------------------------------------------------
	** get context from wlandevice
	** ------------------------------------------------- */
	usbdev=priv->usbdev;
	/* ----------------------------------------------------
	** check which kind of command was issued...
	** ------------------------------------------------- */
	priv->usbout.cmd=cmd;
	priv->usbout.status=0;
	if (cmd==ACX1xx_CMD_INTERROGATE) {
		priv->usbout.u.rridreq.rid=((memmap_t *)pdr)->type;
		priv->usbout.u.rridreq.frmlen=paramlen-4;
		blocklen=8;
		switch (priv->usbout.u.rridreq.rid) {
			case ACX1xx_IE_SCAN_STATUS:skipridheader=1;break;
		}
		if (skipridheader) acklen=paramlen;
		else acklen=4+paramlen;
		acxlog(L_XFER,"sending interrogate: cmd=%d status=%d rid=%d frmlen=%d\n",priv->usbout.cmd,priv->usbout.status,priv->usbout.u.rridreq.rid,priv->usbout.u.rridreq.frmlen);
	} else if (cmd==ACX1xx_CMD_CONFIGURE) {
		priv->usbout.u.wridreq.rid=((memmap_t *)pdr)->type;
		priv->usbout.u.wridreq.frmlen=paramlen-4;
		memcpy(priv->usbout.u.wridreq.data,&(((memmap_t *)pdr)->m),paramlen-4);
		blocklen=paramlen+4;
	} else if ((cmd==ACX1xx_CMD_ENABLE_RX)||(cmd==ACX1xx_CMD_ENABLE_TX)||(cmd==ACX1xx_CMD_SLEEP)) {
		priv->usbout.u.rxtx.data=1;		/* just for testing */
		blocklen=5;
	} else {
		/* ----------------------------------------------------
		** All other commands (not thoroughly tested)
		** ------------------------------------------------- */
		if ((pdr)&&(paramlen>0)) memcpy(priv->usbout.u.wmemreq.data,pdr,paramlen);
		blocklen=paramlen;
	}
	/* ----------------------------------------------------
	** Obtain the I/O pipes
	** ------------------------------------------------- */
	outpipe=usb_sndctrlpipe(usbdev,0);      /* default endpoint for ctrl-transfers: 0 */
	inpipe =usb_rcvctrlpipe(usbdev,0);      /* default endpoint for ctrl-transfers: 0 */
#ifdef ACX_DEBUG
	acxlog(L_XFER,"sending USB control msg (out) (blocklen=%d)\n",blocklen);
	if (debug&L_DATA) acx100_dump_bytes(&(priv->usbout),blocklen);
#endif
	/* --------------------------------------
	** fill setup packet and control urb
	** ----------------------------------- */
	FILL_SETUP_PACKET(priv->usb_setup,USB_TYPE_VENDOR|USB_DIR_OUT,ACX100_USB_UNKNOWN_REQ1,0,0,blocklen)
	usb_fill_control_urb(priv->ctrl_urb,usbdev,outpipe,priv->usb_setup,&(priv->usbout),blocklen,acx100usb_control_complete,priv);
	priv->ctrl_urb->timeout=timeout;
	ucode=submit_urb(priv->ctrl_urb, GFP_KERNEL);
	if (ucode!=0) {
		acxlog(L_STD,"ctrl message failed with errcode %d\n",ucode);
		return(NOT_OK);
	}
	/* ---------------------------------
	** wait for request to complete...
	** ------------------------------ */
	while (priv->ctrl_urb->status==-EINPROGRESS) {
		udelay(500);
	}
	/* ---------------------------------
	** check the result
	** ------------------------------ */
	result=priv->ctrl_urb->actual_length;
	acxlog(L_XFER,"wrote=%d bytes (status=%d)\n",result,priv->ctrl_urb->status);
	if (result<0) {
		return(NOT_OK);
	}
	/* --------------------------------------
	** Check for device acknowledge ...
	** -------------------------------------- */
	acxlog(L_XFER,"sending USB control msg (in) (acklen=%d)\n",acklen);
	priv->usbin.status=0; /* delete old status flag -> set to fail */
	FILL_SETUP_PACKET(priv->usb_setup,USB_TYPE_VENDOR|USB_DIR_IN,ACX100_USB_UNKNOWN_REQ1,0,0,acklen)
	usb_fill_control_urb(priv->ctrl_urb,usbdev,inpipe,priv->usb_setup,&(priv->usbin),acklen,acx100usb_control_complete,priv);
	priv->ctrl_urb->timeout=timeout;
	ucode=submit_urb(priv->ctrl_urb, GFP_KERNEL);
	if (ucode!=0) {
		acxlog(L_STD,"ctrl message (ack) failed with errcode %d\n",ucode);
		return(NOT_OK);
	}
	/* ---------------------------------
	** wait for request to complete...
	** ------------------------------ */
	while (priv->ctrl_urb->status==-EINPROGRESS) {
		udelay(500);
	}
	/* ---------------------------------
	** check the result
	** ------------------------------ */
	result=priv->ctrl_urb->actual_length;
	acxlog(L_XFER,"read=%d bytes\n",result);
	if (result<0) {
		FN_EXIT(0,result);
		return(NOT_OK);
	}
	if (priv->usbin.status!=1) {
		acxlog(L_DEBUG,"command returned status %d\n",priv->usbin.status);
	}
	if (cmd==ACX1xx_CMD_INTERROGATE) {
		if ((pdr)&&(paramlen>0)) {
			if (skipridheader) {
				memcpy(pdr,&(priv->usbin.u.rmemresp.data),paramlen-4);
				acxlog(L_XFER,"response frame: cmd=%d status=%d\n",priv->usbin.cmd,priv->usbin.status);
				acxlog(L_DATA,"incoming bytes (%d):\n",paramlen-4);
				if (debug&L_DATA) 
				    acx100_dump_bytes(pdr,paramlen-4);
			}
			else {
				memcpy(pdr,&(priv->usbin.u.rridresp.rid),paramlen);
				acxlog(L_XFER,"response frame: cmd=%d status=%d rid=%d frmlen=%d\n",priv->usbin.cmd,priv->usbin.status,priv->usbin.u.rridresp.rid,priv->usbin.u.rridresp.frmlen);
				acxlog(L_DATA,"incoming bytes (%d):\n",paramlen);
				if (debug&L_DATA)
				    acx100_dump_bytes(pdr,paramlen);
			}
		}
	}
	FN_EXIT(0,OK);
	return(OK);
}
#endif


/*****************************************************************************
 *
 * The Upper Layer
 *
 ****************************************************************************/

static short CtlLength[0x16] = {
	0,
	ACX100_IE_ACX_TIMER_LEN,
	ACX1xx_IE_POWER_MGMT_LEN,
	ACX1xx_IE_QUEUE_CONFIG_LEN,
	ACX100_IE_BLOCK_SIZE_LEN,
	ACX1xx_IE_MEMORY_CONFIG_OPTIONS_LEN,
	ACX1xx_IE_RATE_FALLBACK_LEN,
	ACX100_IE_WEP_OPTIONS_LEN,
	ACX1xx_IE_MEMORY_MAP_LEN, /*	ACX1xx_IE_SSID_LEN, */
	0,
	ACX1xx_IE_ASSOC_ID_LEN,
	0,
	ACX1xx_IE_CONFIG_OPTIONS_LEN,
	ACX1xx_IE_FWREV_LEN,
	ACX1xx_IE_FCS_ERROR_COUNT_LEN,
	ACX1xx_IE_MEDIUM_USAGE_LEN,
	ACX1xx_IE_RXCONFIG_LEN,
	0,
	0,
	ACX1xx_IE_FIRMWARE_STATISTICS_LEN,
	0,
	ACX1xx_IE_FEATURE_CONFIG_LEN
	
	};

static short CtlLengthDot11[0x14] = {
	0,
	ACX1xx_IE_DOT11_STATION_ID_LEN,
	0,
	ACX100_IE_DOT11_BEACON_PERIOD_LEN,
	ACX1xx_IE_DOT11_DTIM_PERIOD_LEN,
	ACX1xx_IE_DOT11_SHORT_RETRY_LIMIT_LEN,
	ACX1xx_IE_DOT11_LONG_RETRY_LIMIT_LEN,
	ACX100_IE_DOT11_WEP_DEFAULT_KEY_LEN,
	ACX1xx_IE_DOT11_MAX_XMIT_MSDU_LIFETIME_LEN,
	0,
	ACX1xx_IE_DOT11_CURRENT_REG_DOMAIN_LEN,
	ACX1xx_IE_DOT11_CURRENT_ANTENNA_LEN,
	0,
	ACX1xx_IE_DOT11_TX_POWER_LEVEL_LEN,
	ACX1xx_IE_DOT11_CURRENT_CCA_MODE_LEN,
	ACX1xx_IE_DOT11_ED_THRESHOLD_LEN,
	ACX1xx_IE_DOT11_WEP_DEFAULT_KEY_SET_LEN,
	0,
	0,
	0
};

/*----------------------------------------------------------------
* acx100_configure
*
*
* Arguments:
*
* Returns:
*	1 = success
*	0 = failure
* Side effects:
*
* Call context:
*
* STATUS: FINISHED
*
* Comment:
*
*----------------------------------------------------------------*/
int acx100_configure(wlandevice_t *priv, void *pdr, short type)
{
	UINT16 len, offs = 0;

	/* TODO implement and check other acx111 commands */
	if(priv->chip_type == CHIPTYPE_ACX111 &&
		(type == ACX1xx_IE_DOT11_CURRENT_ANTENNA  
		 || type == ACX1xx_IE_DOT11_ED_THRESHOLD
		 /*|| type == ACX1xx_IE_DOT11_CURRENT_REG_DOMAIN*/
		 || type == ACX1xx_IE_DOT11_CURRENT_CCA_MODE)) {
		acxlog(L_INIT, "Configure Command 0x%02X not supported under acx111 (yet)\n", type);

		return NOT_OK;
	}

	if (type<0x1000)
		len=CtlLength[type];
	else
		len=CtlLengthDot11[type-0x1000];

	if (unlikely(len==0)) {
		acxlog(L_DEBUG,"WARNING: ENCOUNTERED ZEROLENGTH RID (%x)\n",type);
	}
	acxlog(L_XFER,"configuring: type(rid)=0x%X len=%d\n",type,len);
	
	((memmap_t *)pdr)->type = cpu_to_le16(type);
#if (WLAN_HOSTIF==WLAN_USB)
	((memmap_t *)pdr)->length = 0; /* FIXME: is that correct? */
	offs = 0; /* FIXME: really?? */
#else
	((memmap_t *)pdr)->length = cpu_to_le16(len);
	offs = 4;
#endif
	return acx100_issue_cmd(priv, ACX1xx_CMD_CONFIGURE, pdr, len + offs, 5000);
}

/*----------------------------------------------------------------
* acx100_configure_length
*
*
* Arguments:
*
* Returns:
*
* Side effects:
*
* Call context:
*
* STATUS: FINISHED
*
* Comment:
*
*----------------------------------------------------------------*/
inline int acx100_configure_length(wlandevice_t *priv, void *pdr, short type, short len)
{
	((memmap_t *)pdr)->type = cpu_to_le16(type);
#if (WLAN_HOSTIF==WLAN_USB)
	((memmap_t *)pdr)->length = 0; /* FIXME: is that correct? */
#else
	((memmap_t *)pdr)->length = cpu_to_le16(len);
#endif
	return acx100_issue_cmd(priv, ACX1xx_CMD_CONFIGURE, pdr, 
		len + 4, 5000);
}

/*----------------------------------------------------------------
* acx100_interrogate
*
*
* Arguments:
*
* Returns:
*	OK = success
*	NOT_OK = failure
*
* Side effects:
*
* Call context:
*
* STATUS: FINISHED
*
* Comment:
*
*----------------------------------------------------------------*/
int acx100_interrogate(wlandevice_t *priv, void *pdr, short type)
{
	UINT16 len;

	if (type<0x1000)
		len=CtlLength[type];
	else
		len=CtlLengthDot11[type-0x1000];

	((memmap_t *)pdr)->type = cpu_to_le16(type);
#if (WLAN_HOSTIF==WLAN_USB)
	((memmap_t *)pdr)->length = 0; /* FIXME: is that correct? */
#else
	((memmap_t *)pdr)->length = cpu_to_le16(len);
#endif

	acxlog(L_XFER,"interrogating: type(rid)=0x%X len=%d\n",type,len);

	return acx100_issue_cmd(priv, ACX1xx_CMD_INTERROGATE, pdr,
		len + 4, 5000);
}


/*****************************************************************************
 * 
 * MAC Address Stuff
 *
 ****************************************************************************/

/*----------------------------------------------------------------
* acx100_is_mac_address_zero
*
*
* Arguments:
*
* Returns:
*
* Side effects:
*
* Call context:
*
* STATUS: FINISHED
*
* Comment:
*
*----------------------------------------------------------------*/
inline int acx100_is_mac_address_zero(mac_t *mac)
{
	if ((mac->vala == 0) && (mac->valb == 0)) {
		return OK;
	}
	return NOT_OK;
}

/*----------------------------------------------------------------
* acx100_is_mac_address_equal
*
*
* Arguments:
*
* Returns:
*
* Side effects:
*
* Call context:
*
* STATUS: FINISHED
*
* Comment:
*
*----------------------------------------------------------------*/
inline int acx100_is_mac_address_equal(UINT8 *one, UINT8 *two)
{
	return memcmp(one, two, ETH_ALEN);
}

/*----------------------------------------------------------------
* acx100_is_mac_address_group
*
*
* Arguments:
*
* Returns:
*
* Side effects:
*
* Call context:
*
* STATUS: FINISHED
*
* Comment:
*
*----------------------------------------------------------------*/
inline UINT8 acx100_is_mac_address_group(mac_t *mac)
{
	return mac->vala & 1;
}

/*----------------------------------------------------------------
* acx100_is_mac_address_directed
*
*
* Arguments:
*
* Returns:
*
* Side effects:
*
* Call context:
*
* STATUS: FINISHED
*
* Comment:
*
*----------------------------------------------------------------*/
inline UINT8 acx100_is_mac_address_directed(mac_t *mac)
{
	if (mac->vala & 1) {
		return NOT_OK;
	}
	return OK;
}

/*----------------------------------------------------------------
* acx100_is_mac_address_broadcast
*
*
* Arguments:
*
* Returns:
*
* Side effects:
*
* Call context:
*
* STATUS: FINISHED
*
* Comment:
*
*----------------------------------------------------------------*/
inline int acx100_is_mac_address_broadcast(const UINT8 *const address)
{
	unsigned char bcast_addr[ETH_ALEN] = { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff };

	if (memcmp(address, bcast_addr, ETH_ALEN) == 0)
		return OK;

	/* IPv6 broadcast address */
	if ((address[0] == 0x33) && (address[1] == 0x33))
		return OK;

	return NOT_OK;
}

/*----------------------------------------------------------------
* acx100_is_mac_address_multicast
*
*
* Arguments:
*
* Returns:
*
* Side effects:
*
* Call context:
*
* STATUS: FINISHED
*
* Comment:
*
*----------------------------------------------------------------*/
inline int acx100_is_mac_address_multicast(mac_t *mac)
{
	if (mac->vala & 1) {
		if ((mac->vala == 0xffffffff) && (mac->valb == 0xffff))
			return NOT_OK;
		else
			return OK;
	}
	return NOT_OK;
}

/*----------------------------------------------------------------
* acx100_log_mac_address
*
*
* Arguments:
*
* Returns:
*
* Side effects:
*
* Call context:
*
* STATUS: NEW
*
* Comment:
*
*----------------------------------------------------------------*/
void acx100_log_mac_address(int level, UINT8 *mac)
{
	acxlog(level, "%02X:%02X:%02X:%02X:%02X:%02X",
			mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);
}

/*----------------------------------------------------------------
* acx100_power_led
*
*
* Arguments:
*
* Returns:
*
* Side effects:
*
* Call context:
*
* STATUS: NEW
*
* Comment:
*
*----------------------------------------------------------------*/
void acx100_power_led(wlandevice_t *priv, UINT8 enable)
{
	if(priv->chip_type == CHIPTYPE_ACX111) {
		/* acx111 makes all led handling itself */
		return;
	}
	if (enable)
		acx100_write_reg16(priv, priv->io[IO_ACX_GPIO_OE], 
			acx100_read_reg16(priv, priv->io[IO_ACX_GPIO_OE]) & ~0x0800);
	else
		acx100_write_reg16(priv, priv->io[IO_ACX_GPIO_OE], 
			acx100_read_reg16(priv, priv->io[IO_ACX_GPIO_OE]) | 0x0800);
}

